import pinttr
from pinttr import UnitGenerator

ureg = pinttr.get_unit_registry()


def test_unit_generator():
    """
    Unit tests for :class:`pinttr._generator.UnitGenerator()`.
    """
    # Calling it returns its registered units
    g_length = UnitGenerator(ureg.m)
    assert g_length() == ureg.m

    # Changing its internal state updates returned units
    g_length.units = ureg.km
    assert g_length() == ureg.km

    # We can nest them
    g_length.units = ureg.m
    g_time = UnitGenerator(ureg.s)
    g_speed = UnitGenerator(lambda: g_length() / g_time())
    assert g_speed() == ureg.m / ureg.s

    # Nested generators also update dynamically
    g_length.units = ureg.km
    assert g_speed() == ureg.km / ureg.s

    # Override temporarily updates the internal units
    g_length.units = ureg.m
    with g_length.override(ureg.km):
        assert g_length() == ureg.km
        assert g_speed() == ureg.km / ureg.s
    assert g_length() == ureg.m
    assert g_speed() == ureg.m / ureg.s

    # Also works if passed a string
    g_length.units = ureg.m
    with g_length.override("km"):
        assert g_length() == ureg.km
        assert g_speed() == ureg.km / ureg.s
    assert g_length() == ureg.m
    assert g_speed() == ureg.m / ureg.s

    # Still works if passing a string to override a composed generator
    with g_speed.override("mile/hour"):
        assert g_speed() == ureg.mile / ureg.hour
    assert g_speed() == ureg.m / ureg.s

    # Also works if passed a generator
    g_length.units = ureg.m
    with g_length.override(g_time):
        assert g_length() == ureg.s
    assert g_length() == ureg.m

    # Also updates generators created from callables
    with g_length.override(ureg.km), g_speed.override(ureg.mile / ureg.hour):
        assert g_speed() == ureg.mile / ureg.hour
        assert g_length() == ureg.km
    assert g_speed() == ureg.m / ureg.s
    assert g_length() == ureg.m



import pint
import pytest

import pinttr
from pinttr import interpret_units


def test_interpret_units():
    """
    Unit tests for :func:`pinttrs.interpret_units`.
    """
    # We'll use the default unit registry
    ureg = pinttr.get_unit_registry()

    # fmt: off
    # Normal operation: units are applied and the '_units' field is removed
    assert interpret_units({"a": 1.0, "a_units": "m"}) == {"a": 1.0 * ureg.m}
    # Also works if the key of the magnitude field is an empty string
    assert interpret_units({"": 1.0, "_units": "m"}) == {"": 1.0 * ureg.m}
    # Also works if the magnitude field key is '_units'
    assert interpret_units({
        "_units": 1.0,
        "_units_units": "m"
    }) == {
        "_units": 1.0 * ureg.m
    }

    # If a unit field has no associated magnitude, nothing changes
    assert interpret_units({"a_units": 1.,}) == {"a_units": 1.}
    assert interpret_units({"_units": "m",}) == {"_units": "m"}
    # fmt: on

    # If inplace is False, the dict is not modified
    d = {"a": 1.0, "a_units": "m"}
    assert interpret_units(d) != d
    # If inplace is True, the dict is modified
    interpret_units(d, inplace=True)
    assert d == {"a": 1.0 * ureg.m}

    # Corner cases
    # -- If magnitude entry is already a Pint quantity, conversion is performed ...
    d = interpret_units({"a": 1.0 * ureg.m, "a_units": "km"})
    assert d == {"a": 1.0 * ureg.m}
    assert d["a"].units == ureg.km
    # -- ... and will fail if incompatible units are used
    with pytest.raises(pint.DimensionalityError):
        interpret_units({"a": 1.0 * ureg.s, "a_units": "m"})



import pint
import pytest

import pinttr


def test_set_unit_registry():
    """
    Unit tests for :func:`pinttr._interpret.interpret_units`.
    """
    # The default unit registry is a pint.UnitRegistry instance
    assert isinstance(pinttr.get_unit_registry(), pint.UnitRegistry)
    ureg = pint.UnitRegistry()
    assert ureg is not pinttr.get_unit_registry()

    # We can change it to the value we like
    pinttr.set_unit_registry(ureg)
    assert ureg is pinttr.get_unit_registry()

    # But it must be a pint.UnitRegistry
    with pytest.raises(TypeError):
        pinttr.set_unit_registry(None)



import attr
import pint
import pytest

import pinttr
from pinttr._metadata import MetadataKey
from pinttr.exceptions import UnitsError

ureg = pint.UnitRegistry()


def test_attrib_metadata():
    """
    Unit tests for :func:`pinttrs._make.attrib` (metadata checks on produced
    attribute specifications).
    """
    # If 'units' argument is not passed, behaviour is similar to that of attr.ib()
    field_no_quantity = pinttr.ib(default=ureg.Quantity(0, "m"))
    assert MetadataKey.UNITS not in field_no_quantity.metadata

    # Units are wrapped into generators and registered as field metadata
    field_distance = pinttr.ib(units=ureg.m)
    assert field_distance.metadata[MetadataKey.UNITS]() == ureg.m

    field_angle = pinttr.ib(units=ureg.deg)
    assert field_angle.metadata[MetadataKey.UNITS]() == ureg.deg

    # Units specified with generators  are directly registered as metadata
    ugen = pinttr.UnitGenerator(ureg.m)
    field_distance = pinttr.ib(units=ugen)
    assert field_distance.metadata[MetadataKey.UNITS]() == ureg.m

    # Units registered with a generator can be overridden
    with ugen.override(ureg.s):
        assert field_distance.metadata[MetadataKey.UNITS]() == ureg.s
    assert field_distance.metadata[MetadataKey.UNITS]() == ureg.m

    # If 'units' argument is not a pint.Unit or a callable returning a pint.Unit, raise
    with pytest.raises(TypeError):
        pinttr.ib(units="km")


def test_attrib_converter_validator():
    """
    Unit tests for :func:`pinttrs._make.attrib` (converter and validator).
    """
    ugen = pinttr.UnitGenerator(ureg.m)

    # If no converter is defined, automatic unit conversion and validation is added
    @attr.s
    class MyClass:
        field = pinttr.ib(default=None, units=ugen)

    # Default set to None makes converter optional
    assert MyClass().field is None
    # Automatic unit conversion is performed
    assert MyClass(1.0).field == 1.0 * ureg.m
    # If a generator was used to fetch units at runtime, updating the conversion units is possible
    ugen.units = ureg.km
    assert MyClass(1.0).field == 1.0 * ureg.km
    # We can even change dimensionality
    ugen.units = ureg.s
    assert MyClass(1.0).field == 1.0 * ureg.s

    # If we use a pint.Quantity to init our field, it should pass if units are compatible
    ugen.units = ureg.m
    assert MyClass(1.0 * ureg.m).field == 1.0 * ureg.m
    # And it should raise if units are not compatible
    ugen.units = ureg.s
    with pytest.raises(UnitsError):
        MyClass(1.0 * ureg.m).field

    # With defaults, we should also have automatic conversion and validation upon setting field
    ugen.units = ureg.m
    a = MyClass(1.0)
    assert a.field == 1.0 * ureg.m
    a.field = 1.0 * ureg.km
    assert a.field == 1.0 * ureg.km
    a.field = 1.0
    assert a.field == 1.0 * ureg.m
    with pytest.raises(UnitsError):
        a.field = 1.0 * ureg.s



import enum

import pytest

import pinttr
from pinttr import UnitContext
from pinttr import UnitGenerator

ureg = pinttr.get_unit_registry()


class PhysicalQuantity(enum.Enum):
    LENGTH = "length"
    SPEED = "speed"
    TIME = "time"


def test_unit_context_init():
    """
    Unit tests for :meth:`pinttr._context.UnitContext()`.
    """
    # Empty initialisation doesn't raise
    UnitContext()

    # Init from str: UnitGenerator map
    unit_context = UnitContext({"length": UnitGenerator(ureg.m)})
    assert unit_context.registry == {"length": UnitGenerator(ureg.m)}

    # Init from str: Unit map
    unit_context = UnitContext({"length": ureg.m})
    assert unit_context.registry == {"length": UnitGenerator(ureg.m)}

    # Init from str: str map
    unit_context = UnitContext({"length": "m"}, interpret_str=True)
    assert unit_context.registry == {"length": UnitGenerator(ureg.m)}

    # Init from str: str map: fails if interpret_str is not True
    with pytest.raises(TypeError):
        UnitContext({"length": "m"})

    # Init from Enum: UnitGenerator map
    unit_context = UnitContext(
        {"length": UnitGenerator(ureg.m)}, key_converter=PhysicalQuantity
    )
    assert unit_context.registry == {PhysicalQuantity.LENGTH: UnitGenerator(ureg.m)}

    # Init fails if dict values are not str, pint.Unit or UnitGenerator
    with pytest.raises(TypeError):
        UnitContext({"length": 1.0})


def test_unit_context_getters():
    # Unit generators are evaluated
    unit_context = UnitContext(
        {"length": ureg.m, "time": ureg.s}, key_converter=PhysicalQuantity
    )
    assert unit_context.get("length") == ureg.m
    assert unit_context.get("time") == ureg.s

    # Square brackets can be used to access the getter
    assert unit_context["length"] == unit_context.get("length")
    assert unit_context["time"] == unit_context.get("time")

    # Raise if key is unregistered
    with pytest.raises(KeyError):
        unit_context.get("speed")  # Still, the key must pass conversion

    # More complex unit generators are also evaluated
    unit_context.register(
        "speed",
        UnitGenerator(lambda: unit_context.get("length") / unit_context.get("time")),
    )
    assert unit_context.get("speed") == ureg.m / ureg.s

    # Dictionary expansion returns correct keys and values
    assert unit_context.get_all() == {
        PhysicalQuantity.LENGTH: ureg.m,
        PhysicalQuantity.TIME: ureg.s,
        PhysicalQuantity.SPEED: ureg.m / ureg.s,
    }

    # Square brackets can be used as an alias to register()
    unit_context["length"] = ureg.km
    assert unit_context.get("length") == ureg.km


def test_unit_context_override():
    unit_context = UnitContext(key_converter=PhysicalQuantity)
    unit_context.update(
        {
            "length": ureg.m,
            "time": ureg.s,
            "speed": UnitGenerator(
                lambda: unit_context.get("length") / unit_context.get("time")
            ),
        }
    )

    # Override with dict propagates to stored generators
    with unit_context.override(
        {PhysicalQuantity.LENGTH: ureg.km, PhysicalQuantity.TIME: ureg.h}
    ):
        assert unit_context.get("length") == ureg.km
        assert unit_context.get("time") == ureg.h
        assert unit_context.get("speed") == ureg.Unit("km/h")
    assert unit_context.get("length") == ureg.m
    assert unit_context.get("time") == ureg.s
    assert unit_context.get("speed") == ureg.Unit("m/s")

    # Override with kwargs behaves similarly
    with unit_context.override(length="km", time="h"):
        assert unit_context.get("length") == ureg.km
        assert unit_context.get("time") == ureg.h
        assert unit_context.get("speed") == ureg.Unit("km/h")
    assert unit_context.get("length") == ureg.m
    assert unit_context.get("time") == ureg.s
    assert unit_context.get("speed") == ureg.Unit("m/s")

    # Override with something else than a dict or kwargs fails
    with pytest.raises(TypeError):
        with unit_context.override(1.0):
            pass


def test_unit_context_deferred():
    unit_context = UnitContext(key_converter=PhysicalQuantity)
    unit_context.update(
        {
            "length": ureg.m,
            "time": ureg.s,
            "speed": UnitGenerator(
                lambda: unit_context.get("length") / unit_context.get("time")
            ),
        }
    )

    # deferred() returns a fully functioning UnitGenerator
    ugen = unit_context.deferred("speed")
    assert isinstance(ugen, UnitGenerator)
    with unit_context.override(length="km", time="s"):
        assert ugen() == ureg("km/s")



import pint
import pytest
from pint import DimensionalityError

from pinttr import UnitGenerator
from pinttr.util import always_iterable, ensure_units, units_compatible

ureg = pint.UnitRegistry()


def test_always_iterable():
    """
    Unit tests for :func:`pinttrs.util.always_iterable`.
    More tests on original code [https://github.com/more-itertools].
    """
    # Utility function to check for success
    def is_iterable(value):
        try:
            iter(value)
            return True
        except TypeError:
            return False

    # Non-iterable becomes iterable
    assert is_iterable(always_iterable(1))
    assert is_iterable(always_iterable(None))
    # Iterable remains iterable
    assert is_iterable(always_iterable([1, 1]))
    # String is packed in an iterable
    assert list(always_iterable("abc")) == ["abc"]


def test_units_compatible():
    """
    Unit tests for :func:`pinttrs.util.units_compatible`.
    """
    # Units with the same dimension are compatible
    assert units_compatible(ureg.m, ureg.km)
    assert units_compatible(ureg.m, ureg.mile)
    assert units_compatible(ureg.m / ureg.km, ureg.dimensionless)
    assert units_compatible(ureg.Unit("kg * m/s^2"), ureg.Unit("N"))

    # Exception: angles are not compatible with dimensionless
    assert not units_compatible(ureg.rad, ureg.dimensionless)
    assert not units_compatible(ureg.sr, ureg.dimensionless)
    assert not units_compatible(ureg.sr, ureg.rad)


def test_ensure_units():
    """
    Unit tests for :func:`pinttrs.converters.ensure_units`.
    """
    # Units are applied to unitless values
    assert ensure_units(100, ureg.km) == ureg.Quantity(100, "km")
    # Default behaviour does not convert quantity values
    assert ensure_units(ureg.Quantity(100, "m"), ureg.km) == ureg.Quantity(100, "m")
    # If a quantity value with incompatible units is passed, does not raise
    ensure_units(ureg.Quantity(100, "m"), ureg.s)

    # If conversion is requested, it is performed
    assert ensure_units(
        ureg.Quantity(100, "m"), ureg.km, convert=True
    ) == ureg.Quantity(0.1, "km")
    # If a quantity value with incompatible units is passed and conversion is requested, raise
    with pytest.raises(DimensionalityError):
        ensure_units(ureg.Quantity(100, "m"), ureg.s, convert=True)

    # If we pass a unit generator, it is evaluated
    u = UnitGenerator(ureg.km)
    assert ensure_units(100, u) == ureg.Quantity(100, "km")
    assert ensure_units(100, UnitGenerator(ureg.s)) == ureg.Quantity(100, "s")

    # If we don't use a pint.Unit or a UnitGenerator to specify units, raise
    with pytest.raises(TypeError):
        ensure_units(100, "km")


import attr
import pint
import pytest

import pinttr
from pinttr.exceptions import UnitsError
from pinttr.validators import has_compatible_units


def test_has_compatible_units():
    """
    Unit tests for :func:`pinttr.validators.has_compatible_units`.
    """
    ureg = pint.UnitRegistry()

    @attr.s
    class MyClass:
        length = pinttr.ib(
            default=0.0 * ureg.m,
            units=ureg.m,
            validator=has_compatible_units,
            converter=None,
        )
        angle = pinttr.ib(
            default=0.0 * ureg.deg,
            units=ureg.deg,
            validator=has_compatible_units,
            converter=None,
        )

    # Validation passes if units have the same dimensionality
    MyClass(length=1.0 * ureg.km)
    MyClass(length=1.0 * ureg.mile)
    MyClass(angle=1.0 * ureg.rad)

    # Validation fails even if units have the same dimensionality but represent different quantities
    with pytest.raises(UnitsError):
        MyClass(angle=1.0 * ureg.dimensionless)

    # Validation fails if value has no units
    with pytest.raises(UnitsError):
        MyClass(angle=1.0)
